from common import ProcessingStage, MultiStage
from util import BroadcastQueue
import inspect

class Fanin(MultiStage):
    """
    This merges all of the outputs of the child stages to a single queue.

    Syntax::

        with Fanin():
            Udp()
            Zeromq()
    """

    def setup(self, q):
        self.output = q
        for s in self.stages:
            s.setup(self.output)
        return None # fanin has no inputs

class Sequence(MultiStage):
    """
    This connects the output of the preceding stage to the input of the next,
    and so on, so the event is processed by each stage one after the other, in
    order.

    Syntax::

        with Sequence():
            Mutate()
            Mutate()
            ...
    """

    def setup(self, q):
        self.output = q
        # this is setup backwards, so the output from the current is
        # connected to the input of the successor, and the input of the current
        # will be connected to the output of the predecessor.
        for s in reversed(self.stages):
            q = s.setup(q)
        self.input = q
        return self.input

class Fanout(MultiStage):
    """
    This enqueues the event onto multiple input queues in parallel.

    Syntax::

        with Fanout():
            Log()
            Elasticsearch()
            Mongodb()
            ...
    """

    def setup(self, q):
        self.output = q
        queues = []
        for s in self.stages:
            queues.append(s.setup(q))
        self.input = BroadcastQueue(queues)
        return self.input

from contextlib import contextmanager

class DefaultDictProxy(object):
    """Proxies attribute request to dict, missing keys default to None"""
    def __init__(self, d):
        self.d = d

    def __getattr__(self, k):
        # properties on event (eg .tags) take preference over elements
        return self[k]

    def __getitem__(self, k):
        if hasattr(self.d, k):
            return getattr(self.d, k)
        else:
            return self.d.get(k)

class Switch(ProcessingStage, MultiStage):
    """Branch flow based on a condition.

    The cases are specified using this syntax. The condition may be a lambda
    expression or code string::

        with Switch() as case:
            with case(lambda ev: ev.field == 'value'):
                Json()
            with case('field2 == "value2"'):
                Mutate()
            with case.default:
                Regex(regex='abc')
    """

    def __init__(self, on_error='reject'):
        super(Switch, self).__init__(on_error=on_error)
        self.cases = []

    @contextmanager
    def __call__(self, condition):
        if isinstance(condition, str):
            # python code as string
            code = compile(condition, 'string', 'eval')
            condition = lambda ev: eval(code, {}, ev)

        br = Sequence()
        self.cases.append((condition, br))
        with br:
            yield

    @property
    def default(self):
        return self(lambda x: True)

    def setup(self, q):
        # separate queue for the incoming condition,
        # and fan in on the individual pipelines.
        for t, br in self.cases:
            br.setup(q)
        return super(Switch, self).setup(q)

    def start(self):
        # call both multistage to start the branch and ProcessingStage greenlet
        ProcessingStage.start(self)
        MultiStage.start(self)

    def stop(self):
        # stop the branch and our greenlet
        ProcessingStage.stop(self)
        MultiStage.stop(self)

    def process(self, event):
        # pass the event into the sub-queue for the applicable pipeline
        ret = True
        for case, br in self.cases:
            if case(DefaultDictProxy(event)):
                br.input.put(event)
                ret = False
                break

        # if no condition handles it, pass straight on (True)
        return ret

class If(ProcessingStage, MultiStage):
    """
    Conditionally execute stages.

    The syntax is as follows. The condition may be a lambda expression or code
    string::

        with If('field==1'):
            Json()
    """

    def __init__(self, condition, on_error='reject'):
        super(If, self).__init__(on_error=on_error)
        if isinstance(condition, str):
            # python code as string
            self.condition_text = condition
            code = compile(condition, 'string', 'eval')
            condition = lambda ev: eval(code, {}, ev)
        else:
            self.condition_text = inspect.getsource(condition)
        self.condition = condition

    # configuration contexts
    def __enter__(self):
        super(If, self).__enter__()
        self.branch = Sequence()
        return self.branch.__enter__()

    def __exit__(self, *exc_info):
        self.branch.__exit__()
        return super(If, self).__exit__()

    def setup(self, q):
        # pass output queue to the branch
        self.branch.setup(q)
        return super(If, self).setup(q)

    def start(self):
        # call both multistage to start the branch and ProcessingStage greenlet
        ProcessingStage.start(self)
        MultiStage.start(self)

    def stop(self):
        # stop the branch and our greenlet
        ProcessingStage.stop(self)
        MultiStage.stop(self)

    def process(self, event):
        # pass the event into the sub-queue for the applicable pipeline
        result = self.condition(DefaultDictProxy(event))
        self.logger.debug("Condition: %s evaluated to %s" % (self.condition_text, result))
        if result:
            self.branch.input.put(event)
            # sub-pipeline will dequeue
            return False
        else:
            # if condition False, pass straight on (True)
            return True



from .filter import Filter

class Python(Filter):
    """Call out to a python function for adding custom functionality.

    :param callable function: callable taking the event as an argument

    Example::

        Python(function=lambda ev: ev.count = int(ev.count))

    Alternatively, a function can be passed, for more complex functionality::

        def clean(ev):
            ev.header = ev.header.strip()
            ev.message = ev.message.strip()

        Python(function=clean)
    """

    def __init__(self, function, on_error='reject'):
        super(Python, self).__init__(on_error=on_error)
        self.function = function

    def process(self, event):
        return self.function(event)



import logging
import gevent
from gevent.queue import JoinableQueue
import gevent.monkey
gevent.monkey.patch_thread()
import threading
from context import Context, ContextManager

# Yuck - workaround python 2.7 bug:
# http://stackoverflow.com/questions/13193278/understand-python-threading-bug
threading._DummyThread._Thread__stop = lambda x: 42

class Stage(object):
    """Base class for all stages.

    Common parameters to all stages:
    :param string on_error: 'tag' or 'reject' - determines behaviour on error (default: reject)
    """

    def __init__(self, on_error='reject'):
        self.logger = logging.getLogger(type(self).__name__)
        self.busy = threading.Lock()
        self.on_error = on_error
        self.register()

    def __str__(self):
        return type(self).__name__

    def register(self):
        Context.instance.current().add(self)

    def configure(self):
        """Configure the stage"""
        pass

    def _error(self, event, reason=None):
        if self.on_error == 'tag':
            event.add_tag('error')
            if reason:
                event['message'] = str(reason)
            self.output.put(event)
        else:
            # otherwise ignore
            if isinstance(reason, Exception):
                self.logger.exception('ignoring %s: %s' % (event, reason))
            else:
                self.logger.warn('ignoring %s: %s' % (event, reason))

    # implement these in base classes
    def setup(self, q):
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError


class SpawnedStage(Stage):
    """A stage that is spawned into a separate greenlet, and _run called"""

    def setup(self, q):
        self.output = q
        self.input = JoinableQueue()
        return self.input

    def start(self):
        """Start the stage running in a greenlet."""
        # run each in a greenlet
        self.g = gevent.spawn(self._run)

    def _run(self):
        pass

    def stop(self):
        """Stop the stage from running."""
        # if necessary, avoid to more gracefully handle termination in the stage
        with self.busy:
            self.g.kill()
            self.g.join()
        self.logger.debug('Stopped')

class ProcessingStage(SpawnedStage):
    """A spawned stage, that processes events one by one in process()"""

    def _run(self):
        while True:
            event = self.input.get()

            # block exit whilst processing an event
            with self.busy:
                try:
                    ret = self.process(event)
                    if ret is not False and self.output:
                        self.output.put(event)
                except Exception as ex:
                    self._error(event, ex)

    def process(self, event):
        pass

class MultiStage(Stage, ContextManager):
    """Base class for stages with children"""

    def __init__(self, **kwargs):
        self.stages = []
        super(MultiStage, self).__init__(**kwargs)

    def __str__(self):
        x = ', '.join(str(s) for s in self.stages)
        return '%s(%s)' % (type(self).__name__, x)

    def add(self, stage):
        """Add this stage to the list"""
        self.stages.append(stage)

    def start(self):
        """Spawn the greenlets for all the inputs, filters and outputs."""
        for s in self.stages:
            self.logger.debug('Starting %s' % s)
            s.start()
            self.logger.debug('Started %s' % s)

    def stop(self):
        """Stop all the inputs, filters and outputs."""
        for s in self.stages:
            self.logger.debug('Stopping %s' % s)
            s.stop()
            self.logger.debug('Stopped %s' % s)



import gevent
import os
import glob
import logging
import errno

from ..event import Event
from .input import Input
from gevent.queue import JoinableQueue

class Tail(gevent.Greenlet):
    """Asynchronously tails a file by name, delivering new lines to a queue."""
    def __init__(self, path, queue, statedir):
        super(Tail, self).__init__()
        self.logger = logging.getLogger('Tail')
        self.path = path
        if statedir:
            self.offset_path = os.path.join(statedir, self.path.replace('/', '_') + '.offset')
        else:
            self.offset_path = self.path + '.offset'
        self.queue = queue
        self.fin = None
        self.start()

    def _run(self):
        try:
            self.tail()
        except:
            self._write_state_file()
            if self.fin:
                self.logger.debug('Closed: %s' % self.path)
                self.fin.close()

    def _write_state_file(self):
        with file(self.offset_path, 'w') as fout:
            offset = self.fin.tell()
            self.logger.debug("Writing state file: %s at offset %d" % (self.offset_path, offset))
            print >>fout, offset

    def tail(self):
        self._ensure_open()
        if os.path.exists(self.offset_path):
            offset = file(self.offset_path).read()
            offset = int(offset)
            self.logger.debug("Seeking in %s to %d" % (self.path, offset))
            self.fin.seek(offset)

        last_st = None

        while True:
            try:
                st = os.stat(self.path)
            except OSError as ex:
                # ignore "No such file"
                if ex.errno != errno.ENOENT:
                    raise
                st = None

            if st == last_st:
                gevent.sleep(0.2)
                continue

            # read pending lines
            line = self.fin.readline()
            while line:
                self.queue.put(line)
                line = self.fin.readline()

            # check for file rolling
            if not st or (last_st and st.st_ino != last_st.st_ino):
                # file has rolled, close and open again
                self.logger.debug('Detected roll: %s' % self.path)
                self._ensure_open()
                st = None # restart
            elif st and last_st and st.st_size < last_st.st_size:
                self.logger.debug('Detected truncation: %s' % self.path)
                self._ensure_open()
                st = None # restart

            last_st = st

    def _ensure_open(self):
        if self.fin:
            self.fin.close()
            self.fin = None

        while True:
            try:
                self.fin = file(self.path, 'r')
                self.logger.debug('Opened: %s' % self.path)
                return
            except IOError as ex:
                if ex.errno == errno.ENOENT:
                    gevent.sleep(0.01)
                    continue
                raise

class File(Input):
    """Tails events from a log file on disk.

    Creates events with the field 'data' set to the line received.

    :param string path: path on the file system to the log file(s), wildcards may
      be used to match multiple files.
    :param string statedir: writable directory to store state for files

    Example::

        File(path='/var/log/syslog')
    """

    def __init__(self, path, statedir=None):
        super(File, self).__init__()
        self.path = path
        self.statedir = statedir
        self.tails = []

    def _run(self):
        paths = glob.glob(self.path)
        while not paths:
            gevent.sleep(0.01)
            paths = glob.glob(self.path)
        q = JoinableQueue()

        self.logger.debug('Tailing %s' % ', '.join(paths))
        self.tails = [Tail(p, q, self.statedir) for p in paths]

        while True:
            data = q.get()
            if data:
                if data.endswith('\n'):
                    data = data[0:-1]
                self.logger.debug('Received: %r' % data)
                self.output.put(Event(data=data))
            q.task_done()

    def stop(self):
        for t in self.tails:
            t.kill()
            t.join()
        super(File, self).stop()



import gevent
import gevent.socket as socket
import pickle
import struct
import time

from .output import Output

class Graphite(Output):
    """Upload stats data to a graphite server.

    :param string host: graphite server hostname
    :param string port: graphite server port

    Example::

        Graphite(host='graphite')
    """

    def __init__(self, host='localhost', port=2004):
        super(Graphite, self).__init__()
        self.host = host
        self.port = port
        self._metrics = []
        self._connected = False
        
    def flush(self):
        if not self._connected:
            self.connect()

        # wire protocol is a 4-byte length, followed by pickled representation,
        # see: http://graphite.readthedocs.org/en/1.0/feeding-carbon.html
        payload = pickle.dumps(self._metrics)
        header = struct.pack("!L", len(payload))
        message = header + payload
        self.logger.debug("Flushing %d metrics to graphite %s:%d" % (len(self._metrics), self.host, self.port))
        try:
            self.sock.sendall(message)
        except socket.error:
            self.sock.close()
            self.connect()
            self.sock.sendall(message)
        self._metrics = []

    def connect(self):
        while True:
            try:
                self.sock = socket.socket()
                self.sock.connect((self.host, self.port))
                break
            except socket.error:
                self.logger.warn("Couldn't connect to graphite: %s:%s, retrying in 1s" % (self.host, self.port))
                gevent.sleep(1.0)

        self._connected = True
        
    def process(self, event):
        metric = event.metric
        timestamp = int(time.mktime(event.timestamp.timetuple()))
        for s, value in event.stats.iteritems():
            path = '%s.%s' % (metric, s)
            d = (path, (timestamp, value))
            self._metrics.append(d)

        self.flush()



import gevent
import gevent.socket

from ..event import Event
from .input import Input

class Udp(Input):
    """Receives from a udp port.

    Creates events with the field 'data' set to the packet received.

    :param integer port: listening port

    Example::

        Udp(port=6000)
    """

    def __init__(self, port):
        super(Udp, self).__init__()
        self.port = port
        self.sock = gevent.socket.socket(gevent.socket.AF_INET, gevent.socket.SOCK_DGRAM)
        self.sock.setsockopt(gevent.socket.SOL_SOCKET, gevent.socket.SO_BROADCAST, 1)
        self.sock.bind(('', self.port))

    def _run(self):
        while True:
            data = self.sock.recv(4096)
            self.logger.debug('Received: %r' % data)
            self.output.put(Event(data=data))
            gevent.sleep() # yield for other stages



import tempfile
import shutil
import os

def assertEventEquals(self, expected, actual):
    """Compares two events for equality, ignoring the timestamps."""
    expected = dict(expected)
    del expected['timestamp']
    actual = dict(actual)
    del actual['timestamp']
    self.assertEquals(expected, actual)

class Matcher(object):
    """Helper for test assertions"""
    def __init__(self, compare):
        self.compare = compare
        
    def __eq__(self, other):
        return self.compare(other)
    
    def __repr__(self):
        return repr(self.compare)
        
def any(_):
    """Matches anything"""
    return True

def about(x, p=0):
    fmt = '%%.%df' % p
    return Matcher(lambda t: fmt % x == fmt % t)

def between(x, y):
    return Matcher(lambda t: t > x and t < y)
    
ANY = Matcher(any)

class TempDirectory(object):
    """Temporary directory with context manager support. Inspired by python 3."""

    def __init__(self, suffix='', prefix='tmp', dir=None, change=True):
        self.suffix = suffix
        self.prefix = prefix
        self.dir = dir
        self.change = change

    def __enter__(self):
        self.name = tempfile.mkdtemp(self.suffix, self.prefix, self.dir)
        if self.change:
            self.oldpwd = os.getcwd()
            os.chdir(self.name)
        return self

    def __exit__(self, *excinfo):
        if self.change:
            os.chdir(self.oldpwd)
        shutil.rmtree(self.name)



import time
import gevent

from .output import Output

class Perf(Output):
    """Simple performance counter output.

    :param integer period: interval between reports in seconds

    Example::

        Perf(period=5)
    """
    def __init__(self, period=60):
        super(Perf, self).__init__()
        self.period = period
        self.now = time.time()
        self.count = 0

    def start(self):
        super(Perf, self).start()

        # spawn additional greenlet to do periodic reporting
        gevent.spawn(self._report)

    def _report(self):
        while True:
            gevent.sleep(self.period)

            now = time.time()
            if self.count > 0:
                self.logger.info('%d in %ds (%.1f/s)' % (
                    self.count, now-self.now, self.count/(now-self.now)))
            self.now = now
            self.count = 0

    def process(self, event):
        self.count += 1



from .output import Output
import gevent
import gevent.monkey
gevent.monkey.patch_socket()
import urllib2
import json

class Elasticsearch(Output):
    """Outputs to an elasticsearch index.

    :param string host: elasticsearch host
    :param integer port: elasticsearch port
    :param string index: (required) elasticsearch index. This can be formatted by fields in the event.
    :param string type: (required) elasticsearch type. This can be formatted by fields in the event.

    Example configuration for kibana::

        Mutate(rename={'@timestamp': 'timestamp', '@message': 'message'})
        Elasticsearch(index='logstash-{@timestamp:%Y.%m.%d}', type='event')
    """

    RETRIES = 10

    def __init__(self, index, type, host='localhost', port=9200):
        super(Elasticsearch, self).__init__()
        self.host = host
        self.port = port
        self.index = index
        self.type = type

    def process(self, event):
        data = event.to_json()
        index = event.format(self.index)
        itype = event.format(self.type)
        if not index:
            raise ValueError("index is empty")
        if not itype:
            raise ValueError("type is empty")

        url = 'http://%s:%s/%s/%s/' % (self.host, self.port, index, itype)

        success = False
        delay = 1.0
        for retry in xrange(self.RETRIES):
            try:
                res = urllib2.urlopen(url, data=data)
                result = json.load(res)
                success = result['ok']
                if not success:
                    self.logger.error('Indexing failed: %s' % result)
                break
            except urllib2.HTTPError as ex:
                if ex.getcode() == 400:
                    # Bad Request - do not retry
                    self.logger.error("Bad request: %s, not retrying" % (ex,))
                    break
                else:
                    delay *= 2.0
                    self.logger.warn('Unable to index: %s, retrying in %ds' % (ex, delay))
                    gevent.sleep(delay)
            except urllib2.URLError as ex:
                delay *= 2.0
                self.logger.warn('Unable to index: %s, retrying in %ds' % (ex, delay))
                gevent.sleep(delay)

        if success:
            self.logger.debug('Indexed to elasticsearch: index:%s type:%s id:%s' % (index, itype, result['_id']))



class Context(object):
    def __init__(self):
        self._stack = []

    def push(self, c):
        self._stack.append(c)

    def pop(cls):
        return cls._stack.pop()

    def current(cls):
        return cls._stack[-1]

# Singleton
Context.instance = Context()

class ContextManager(object):
    def __enter__(self):
        Context.instance.push(self)
        return self

    def __exit__(self, *exc_info):
        Context.instance.pop()

    def add(self, stage):
        """Override add to handle child adds"""
        pass

class DummyContext(ContextManager):
    def __init__(self):
        self.stages = []

    def add(self, stage):
        self.stages.append(stage)



import re
import dateutil.parser

from .filter import Filter

class Regex(Filter):
    """Parse a field with a regular expression. The regex named groups
    ``(?P<name>...)`` will be create event fields (overwriting any existing).

    If you extract a 'timestamp' field, this will automatically be parsed as a
    datetime and used as the event timestamp (instead of the default of the time
    received).

    :param string regex: the regular expression
    :param string field: the field to run the regex on (default: data)

    Example::

        Regex(regex='(?P<timestamp>.+) - (?P<message>.+)')
    """

    def __init__(self, regex, field='data', on_error='reject'):
        super(Regex, self).__init__(on_error=on_error)
        self.field = field
        self.regex = re.compile(regex)

    def process(self, event):
        if self.field in event:
            data = event.pop(self.field)
            m = self.regex.search(data)
            if m:
                d = m.groupdict()
                if 'timestamp' in d:
                    # if the originating timestamp is being extracted, convert
                    # it to a datetime
                    d['timestamp'] = dateutil.parser.parse(d['timestamp'])
                self.logger.debug('Matched: %s' % d)
                event.update(d)
            else:
                self._error(event, 'no match')



from .output import Output

class Log(Output):
    """Logging output.

    :param string message: message to log (optional)

    Example::

        Log(message="event:")
    """
    def __init__(self, message=''):
        super(Log, self).__init__()
        self.message = message
        if message:
            self.message += ' '

    def process(self, event):
        self.logger.info('%s%r' % (self.message, event))



import re
import dateutil.parser

from .filter import Filter

class Syslog(Filter):
    """Parse a syslog encoded field.

    This sets the fields:
    
    - timestamp
    - facility
    - severity
    - host
    - program
    - pid
    - message

    :param string field: the field containing the syslog message (default: data)
    :param boolean consume: whether to remove the field after decoding (default: true)

    Example::

        Syslog()
    """

    def __init__(self, field='data', consume=True, on_error='reject'):
        super(Syslog, self).__init__(on_error=on_error)
        self.field = field
        self.consume = consume

    re_syslog = re.compile(r'<(?P<prio>\d+)>(?P<timestamp>.+?) '
        r'(?P<host>\S*) (?P<program>\S+?)'
        r'(\[(?P<pid>\d+)\])?: (?P<message>.+)')
    facilities = ["kernel", "user-level", "mail", "system", "security/authorization",
        "syslogd", "line printer", "network news", "UUCP", "clock", "security/authorization",
        "FTP", "NTP", "log audit", "log alert", "clock", "local0", "local1", "local2",
        "local3", "local4", "local5", "local6", "local7"]
    severities = ["Emergency", "Alert", "Critical", "Error", "Warning", "Notice",
        "Informational", "Debug"]
    months = dict((v, k) for k, v in enumerate(['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']))

    def process(self, event):
        if self.field in event:
            data = event[self.field]

            try:
                fields = self._decode(data)
                self.logger.debug('syslog decoded: %s' % fields)
                event.update(fields)
                if self.consume:
                    del event[self.field]

                return True
            except ValueError as ex:
                self._error(event, ex)
                return False

    def _decode(self, data):
            m = self.re_syslog.match(data)
            if m:
                d = m.groupdict()
                prio = int(d.pop('prio'))
                try:
                    d['facility'] = self.facilities[prio >> 3]
                except IndexError:
                    d['facility'] = 'unknown'
                d['severity'] = self.severities[prio & 7]

                timestamp = d.pop('timestamp')
                d['timestamp'] = dateutil.parser.parse(timestamp)
                return d
            else:
                raise ValueError('invalid syslog')



from .filter import Filter
import logging
import time

from ..statistics import mean, percentile, stddev
from ..util import Periodic, get_path
from ..event import Event
from pprint import pformat

class Stats(Filter):
    """Filter that produces aggregate statistics.

    It will produce:

    - name.count: number of data points
    - name.rate: the data points per second
    - name.mean: mean of data points
    - name.min: minimum data point
    - name.median: median data point
    - name.upper95: 95th% data point
    - name.upper99: 99th% data point
    - name.max: maximum data point
    - name.stddev: standard deviation

    This is emitted as a single event, every period.

    :param integer period: period to report stats, in seconds
    :param map metrics: field names => values. Any fields from the events can
        be formatting into the field names. Values can be an event field, nested
        path to a field (separated by .) and can contain wildcard '*', to indicate
        generating statistics from any numerical fields.
    :param boolean zero: output zero for previously seen metrics (useful to disambiguate
        no activity and output broken)

    Example::

        Stats(metrics={'rails.{controller}.{action}.duration': 'duration'})

    Wildcards can be used to pull out nested structures::

        Stats(metrics={'app.{1}': 'timings.*'})
    """
    def __init__(self, period=5, metrics=None, zero=True):
        super(Stats, self).__init__()
        # configuration
        self.metrics = metrics or {}
        self.zero = zero

        # transient state
        self.timers = {}

        self.last = time.time()
        self.periodic = Periodic(period, self.flush)

    def process(self, event):
        for output, path in self.metrics.iteritems():
            values = get_path(event, path)
            for key, value in values:
                if isinstance(value, (int, float)):
                    self._process_value(event, output, key, value)

    def _process_value(self, event, output, v, value):
        try:
            k = event.format(output, [v], raise_missing=True)
            try:
                # optimise for common case
                self.timers[k].add(value)
            except KeyError:
                self.timers[k] = Timer()
                self.timers[k].add(value)
        except KeyError:
            # event didn't contain all the necessary format keys - ignore
            pass

    def start(self):
        super(Stats, self).start()
        self.periodic.start()

    def stop(self):
        self.periodic.kill()
        self.periodic.join()
        super(Stats, self).stop()

    def flush(self):
        count = 0
        now = time.time()
        # calculate period for precise rate calculation
        period = now - self.last
        for k, timer in self.timers.iteritems():
            stats = timer.stats(period, self.zero)
            if stats:
                count += 1
                if self.output:
                    self.output.put(Event(tags=['stat'], metric=k, stats=stats))
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug('%s: %s' % (k, pformat(stats)))
            timer.reset()
        if count:
            self.logger.debug('Flushed %d stats' % count)
        self.last = now

class Timer(object):
    def __init__(self):
        self.values = []

    def add(self, v):
        # Not much smarts to this. If we're looking at bigger data,
        # then an automatic downsampling would make sense here.
        self.values.append(v)

    def stats(self, period, zero):
        if self.values or zero:
            d = {}
            d['count'] = len(self.values)
            d['rate'] = d['count'] / period
            d['mean'] = mean(self.values)
            self.values.sort()
            d['min'] = percentile(self.values, 0.0)
            d['median'] = percentile(self.values, 0.5)
            d['upper95'] = percentile(self.values, 0.95)
            d['upper99'] = percentile(self.values, 0.99)
            d['max'] = percentile(self.values, 1.0)
            d['stddev'] = stddev(self.values, d['mean'])
            return d
        return None

    def reset(self):
        self.values[:] = []



import urlparse

from .filter import Filter

class Url(Filter):
    """Parse a field with a regular expression. The regex named groups
    ``(?P<name>...)`` will be create event fields (overwriting any existing).

    If you extract a 'timestamp' field, this will automatically be parsed as a
    datetime and used as the event timestamp (instead of the default of the time
    received).

    :param string field: the field to run the regex on (default: data)
    """

    def __init__(self, field='data', on_error='reject'):
        super(Url, self).__init__(on_error=on_error)
        self.field = field

    def process(self, event):
        if self.field in event:
            url = event.pop(self.field)
            url.split('?', 1)
            r = urlparse.urlparse(url)
            event['path'] = r.path
            # for simplicity, ignore repeated parameters
            event.update(dict(urlparse.parse_qsl(r.query)))



import json
from string import Formatter
from datetime import datetime
from pprint import pformat
import dateutil.tz

class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        import datetime
        if isinstance(obj, datetime.datetime):
            if obj.tzinfo is not None:
                # normalize all datetimes to UTC
                obj = obj.astimezone(dateutil.tz.tzutc()).replace(tzinfo=None)

            # isoformat doesn't produce consistent output when microsecond=0
            ms = '.%03dZ' % (obj.microsecond/1000)
            return obj.strftime("%Y-%m-%dT%H:%M:%S") + ms
        else:
            return super(JSONEncoder, self).default(obj)

class DefaultFormatter(Formatter):
    default = ''

    def get_value(self, key, args, kwargs):
        # Try standard formatting, then return 'unknown key'
        try:
            return Formatter.get_value(self, key, args, kwargs)
        except KeyError:
            return self.default

class Event(dict):
    """An event.

    This is the basic unit of communication in logcabin.
    """
    def __init__(self, *args, **kwargs):
        """
        Create an event. A timestamp is automatically generated.
        """
        if 'timestamp' not in kwargs:
            self['timestamp'] = datetime.utcnow()
        super(Event, self).__init__(*args, **kwargs)

    def add_tag(self, value):
        """
        Add a tag to the event.

        >>> ev = Event()
        >>> ev.add_tag('tag1')
        >>> ev.tags
        ['tag1']
        """
        self.setdefault('tags', []).append(value)

    @property
    def tags(self):
        """Get the tags for the event.

        >>> Event().tags
        []
        """
        return self.get('tags', [])

    def __getattr__(self, k):
        """Attribute getter for the dictionary elements, with None as default. Default makes
        safe to missing attributes in conditionals, etc.

        >>> Event().a
        >>> Event(a=2).a
        2
        """
        return self.get(k)

    def __setattr__(self, k, v):
        """Attribute setter for the dictionary elements.

        >>> ev = Event(timestamp=datetime(2013, 1, 1, 1, 2, 3, 45))
        >>> ev.a = 2
        >>> ev.b = ['1']
        >>> str(ev)
        "Event({'a': 2, 'b': ['1'], 'timestamp': datetime.datetime(2013, 1, 1, 1, 2, 3, 45)})"
        """
        self[k] = v

    def to_json(self):
        """Serialize the event to a json string.

        >>> from mock import patch
        >>> with patch('logcabin.event.datetime') as m:
        ...     m.utcnow.side_effect = lambda: datetime(2013, 1, 1, 2, 34, 56, 789012)
        ...     Event(field='x').to_json()
        '{"timestamp":"2013-01-01T02:34:56.789Z","field":"x"}'
        >>> Event(timestamp=datetime(2013, 1, 1, 1, 2, 3, 45)).to_json()
        '{"timestamp":"2013-01-01T01:02:03.000Z"}'
        """
        return json.dumps(self, cls=JSONEncoder, separators=(',', ':'))

    def format(self, fmt, args=None, raise_missing=False):
        """Format the event using string.format notation.

        >>> Event(field='x').format("field={field} missing={missing}")
        'field=x missing='
        >>> Event(field='x').format("field={field} arg1={0} arg2={1}", ['apple', 'pear'])
        'field=x arg1=apple arg2=pear'
        >>> from mock import patch
        >>> with patch('logcabin.event.datetime') as m:
        ...     m.utcnow.side_effect = lambda: datetime(2013, 1, 1, 2, 34, 56, 678912)
        ...     Event(field='x').format("{timestamp:%A %d. %B %Y}")
        'Tuesday 01. January 2013'
        """
        formatter = raise_missing and Formatter() or DefaultFormatter()
        return formatter.vformat(fmt, args, self)

    def __repr__(self):
        """Returns the representation of the object.

        >>> from mock import patch
        >>> with patch('logcabin.event.datetime') as m:
        ...     m.utcnow.side_effect = lambda: datetime(2013, 1, 1, 2, 34, 56)
        ...     repr(Event(field='x'))
        "Event({'field': 'x', 'timestamp': datetime.datetime(2013, 1, 1, 2, 34, 56)})"
        """
        # pprint returns a stable ordering for tests
        return 'Event(%s)' % pformat(dict(self))
